#!/usr/bin/env python2
#
# GEemyle: Eemyle Connector using XML-RPC
# Copyright (C) 2003-2006 Easter-eggs <info@easter-eggs.com>
# Authors:
#      Cyril Lacoux (Yack) <clacoux@easter-eggs.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

__author__  = "Cyril Lacoux [Yack] <clacoux@easter-eggs.com>"
__cvsid__   = "$Id$"
__version__ = "$Revision$"[11:-2]
__all__     = [ "EemyleConnector"]

import base64
import json
import threading
import gobject
import urllib2
import time
import sys
import traceback

import geemyle.config as config
import geemyle.gettext as gettext
_ = gettext._

class EemyleConnectorRESTError(Exception):

    def __init__(self, status_code=None, data=None):
        Exception.__init__(self)
        self.status_code = status_code if status_code else -1
        self.data = data
        self.errors = []
        if isinstance(data, dict):
            if data.get("error"):
                self.errors.append(data["error"])
            if data.get("errors"):
                self.errors.extend(data["errors"])
        if not self.errors:
            self.errors = [_("Unknown error")]

    def __str__(self):
        return "<EemyleConnectorRESTError: [%d] %s>" % (self.status_code, ", ".join(self.errors))

    def __repr__(self):
        return str(self)

class EemyleConnector(threading.Thread, gobject.GObject):
    """
    XML-RPC Connector to Eemyle server
    """

    __hostname = ''
    __wservice = ''
    __username = ''
    __password = ''
    __scheme = ''
    __port = ''

    def __init__(self, verbose=False):
        """
        Initializing some stuff
        """
        threading.Thread.__init__(self, name='Eemyle XML-RPC Connector')
        gobject.GObject.__init__(self)

        # Internal vars
        self.__verbose      = verbose
        self.__root_uri     = None
        self.__do_run       = True
        self.__request_fifo = []

        self.load_config()

        # Start the thread
        self.start()

    def load_config(self):
        """
        Load configuration
        """
        changed = False
        hostname = config.get('eemyle', 'hostname')
        if hostname != self.__hostname:
            self.__hostname = hostname
            changed = True

        keep_u_p = config.get('eemyle', 'keep_u_p')
        username = config.get('eemyle', 'username') if keep_u_p else ''
        if username != self.__username:
            self.__username = username
            changed = True

        password = config.get('eemyle', 'password') if keep_u_p else ''
        if password != self.__password:
            self.__password = password
            changed = True

        wservice = config.get('eemyle', 'wservice') if keep_u_p else ''
        if wservice != self.__wservice:
            self.__wservice = wservice
            changed = True

        port = config.get('eemyle', 'port') if keep_u_p else ''
        if port != self.__port:
            self.__port = str(port)
            changed = True

        scheme = 'https' if config.get('eemyle', 'ssl') else 'http'
        if scheme != self.__scheme:
            self.__scheme = scheme
            changed = True

        if changed:
            self.__build_server()

    def __build_server(self):
        """
        Builds the server depending the connection parameters
        """

        self.__root_uri = '%s://%s:%s' % (self.__scheme, self.__hostname, self.__port)
        if self.__wservice:
            self.__root_uri += "/%s" % self.__wservice

        return True

    def run(self):
        """
        Make the connection and start the thread
        """

        while self.__do_run:
            if self.__request_fifo:
                # Getting the function name and arguments from the fifo
                url, action, kwargs = self.__request_fifo.pop()

                # Sending the request
                result, data = self.__send_request(url, **kwargs)

                # Sending the result to the main process
                self.emit('rest-data-received', action, result, data)

            # Don't eat the CPU
            time.sleep(0.1)

    def __send_request(self, url, **kwargs):
        """
        Sends a REST request to Eemyle API
        """
        # Encoding the arguments
        enc_args = []

        # Calling the function
        try:
            method = kwargs.pop("method", "GET").upper()
            url = "%s/%s" % (self.__root_uri.rstrip("/"), url)
            if self.__verbose:
                sys.stdout.write(
                    '\n%s request on %s %s\n' % (
                        method,
                        url,
                        ("with %s" % str(kwargs)) if kwargs
                        else "without extra args"
                    )
                )

            # Forge request
            request = urllib2.Request(url)
            request.get_method = lambda: method
            request.add_header("User-Agent", "GEemyle/%s" % __version__)
            request.add_header(
                "Authorization",
                "Basic %s" % base64.b64encode('%s:%s' % (self.__username, self.__password))
            )
            if kwargs.get("json"):
                request.add_data(json.dumps(kwargs["json"]))
                request.add_header('Content-Type', 'application/json')

            response = urllib2.urlopen(request)
            status_code = response.getcode()
            data = response.read()

        except urllib2.HTTPError as err:
            result = False
            try:
                data = json.load(err)
            except Exception:
                data = {
                    "success": False,
                    "error": str(err)
                }
            data = EemyleConnectorRESTError(err.code, data)

        except Exception as err:
            # Unknown error
            result   = False
            data     = EemyleConnectorRESTError(
                -1,
                {
                    "success": False,
                    "error": str(err)
                }
            )
            sys.stderr.write(
                '\nError occurred during %s request on %s: %s\n' % (
                    method,
                    url,
                    traceback.format_exc()
                )
            )

        else:
            try:
                data     = json.loads(data)
                result   = data.get("success", False)

                if not result or status_code != 200:
                    sys.stderr.write(
                        '\nError returned on %s request on %s: %s\n' % (
                            method,
                            url,
                            data
                        )
                    )
                    data = EemyleConnectorRESTError(status_code, data)
                elif self.__verbose:
                    sys.stdout.write('\n%s\n' % str(data))
            except ValueError as err:
                sys.stderr.write(
                    '\nError parsing JSON response of %s request on %s: %s\n\n%s\n' % (
                        method,
                        url,
                        data,
                        traceback.format_exc()
                    )
                )
                result   = False
                data     = EemyleConnectorRESTError(
                    -1,
                    {
                        "success": False,
                        "error": _("Failed to parse JSON response of the API: %s") % err
                    }
                )

        return (result, data)

    def send_request(self, url, action, **kwargs):
        """
        Stack a REST request to the internal fifo
        """

        self.__request_fifo.insert(0, [url, action, kwargs])

        return True

    def send_direct_request(self, url, **kwargs):
        """
        Sends a REST request without stacking it
        Directly return the result
        """

        # Sending the request
        return self.__send_request(url, **kwargs)

    def stop(self):

        self.__do_run = False


gobject.type_register(EemyleConnector)
gobject.signal_new('xmlrpc-data-received', EemyleConnector, gobject.SIGNAL_RUN_LAST,
                   gobject.TYPE_NONE, [gobject.TYPE_STRING, gobject.TYPE_BOOLEAN, gobject.TYPE_PYOBJECT])
gobject.signal_new('rest-data-received', EemyleConnector, gobject.SIGNAL_RUN_LAST,
                   gobject.TYPE_NONE, [gobject.TYPE_STRING, gobject.TYPE_BOOLEAN, gobject.TYPE_PYOBJECT])

